package paraphylo;

import java.io.File;
import java.io.IOException;
import java.util.Comparator;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.GnuParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.Options;

import paraphylo.dao.Parameters;
import paraphylo.dataset.MultiGeneFamilySpeciesDataSet;
import paraphylo.parser.Parser;
import paraphylo.solver.cplex.CplexPipeline;

public class Main {

	/**
	 * @param args
	 * @throws IOException 
	 */
	public static void main(String[] args) throws IOException {
		
		long time = System.currentTimeMillis();
		
		Parameters parameters = parseParameters(args);
				
		//Random random = new Random(Long.parseLong("7412470335802172515"));
		MultiGeneFamilySpeciesDataSet ds = null;
		if (parameters.getInputType()==Parameters.COGRAPH_INPUT){
			ds = Parser.parseCographs(parameters);
		}
		else if (parameters.getInputType()==Parameters.TRIPLE_INPUT){
			ds = Parser.parseTriples(parameters);
		}
		else{
			ds = Parser.parseDs(parameters);
		}

		CplexPipeline.solve(ds, parameters);
		
		System.out.println("Computation time: " + (System.currentTimeMillis()-time) + " millisec");

	}
	
	public static Options createOptions(){
		// create the Options
		Option h = new Option( "h", "help", false, "print this message");
		Option i = new Option( "i", "input", true, "input filename" );
		i.setRequired(true);
		Option o = new Option( "o", "output", true, "output filename" );
		o.setRequired(true);
		Option s = new Option( "s", "cograph-size", true, "maximal size of a gene family to be used for the analysis [default=-1]" );
		s.setType(Number.class);
		Option d = new Option( "d", "cograph-distance", true, "maximal cograph edit distance (ratio) of a gene family to be used for the analysis [default=1.0]" );
		d.setType(Number.class);
		Option a = new Option( "a", "cograph-all", false, "compute all optimal cographs" );
		Option w = new Option( "w", "write", false, "write cographs and extreacted triples to file" );
		Option e = new Option( "e", "enrichment", true, "enrich the triple set with orthology data using bitscores between 3 orthologous genes, i.e., a triple is added whenever the ratios between 2 out of the 3 pairs of bitscores is smaler than a threshold [default=0.0]" );
		e.setType(Number.class);
		Option f = new Option( "f", "enrichment-factor", true, "factor for weighting the set of enriched triples, to score the triples obtaind from orthology data different than the triples obtained from paralogy data [default=0.25]" );
		f.setType(Number.class);
		Option b = new Option( "b", "bootstrap", true, "create multiple trees for bootstrapping [default=0]\n\t0 - no bootstrapping\n\t1 - cograph bootstrapping\n\t2 - triple bootstrapping" );
		b.setType(Number.class);
		Option r = new Option( "r", "replications", true, "number of replications for bootstrapping [default=100]" );
		r.setType(Number.class);
		Option u = new Option( "u", "support", true, "minimal support per triple to be considered for the tree reconstruction [default=0.0]" );
		u.setType(Number.class);
		Option t = new Option( "t", "noise-type", true, "add noise to input data [default=0]\n\t0 - no noise\n\t1 - homologous noise\n\t2 - orthologous noise\n\t3 - paralogous noise\n\t4 - xenologous noise" );
		t.setType(Number.class);
		Option p = new Option( "p", "noise-probability", true, "probability of adding noise [default=0.0]" );
		p.setType(Number.class);
		Option c = new Option( "c", "timelimit-cge", true, "time limit in seconds for cograph editing ILP [default=-1]" );
		c.setType(Number.class);
		Option m = new Option( "m", "timelimit-mcs", true, "time limit in seconds for maximal consistent subset ILP [default=-1]" );
		m.setType(Number.class);
		Option l = new Option( "l", "timelimit-lrt", true, "time limit in seconds for least resolved tree ILP [default=-1]" );
		l.setType(Number.class);
		Option q = new Option( "q", "quick-te", false, "disable the polynomial time algorithm for triple extraction and use the slow ILP instead" );
		Option g = new Option( "g", "greedy-cs", false, "do not run the ILP for maximal consistent subset (just do greedy heurictic)" );
		Option n = new Option( "n", "no-ltr", false, "do not run the ILP for least resolved tree (just do BUILD)" );
		Option v = new Option( "v", "verbose", false, "write ILP output to console" );
		
		Options options = new Options();
		options.addOption(h);
		options.addOption(i);
		options.addOption(o);
		options.addOption(s);
		options.addOption(d);
		options.addOption(a);
		options.addOption(w);
		options.addOption(e);
		options.addOption(f);
		options.addOption(b);
		options.addOption(r);
		options.addOption(u);
		options.addOption(t);
		options.addOption(p);
		options.addOption(c);
		options.addOption(m);
		options.addOption(l);
		options.addOption(q);
		options.addOption(g);
		options.addOption(n);
		options.addOption(v);
		
		return options;
	}
	
	public static Parameters parseParameters(String[] args){
		
		// create the command line parser
		CommandLineParser parser = new GnuParser();
		Options options = createOptions();
		Parameters parameters = new Parameters(null);
				
		try {
		    // parse the command line arguments
		    CommandLine line = parser.parse( options, args );
		    
		    if( line.hasOption( "help" ) ) {
		    	showHelp(options);
		    }

		    if( line.hasOption( "input" ) ) {
		    	String input = line.getOptionValue("input");
		    	parameters.setInput(input);
		    	if (input == "" || !new File(input).isFile()){
		    		System.out.println( "ERROR: input file '" + input + "' not found" );
		    		showHelp(options);
		    	}
		    	if (input.endsWith(".cographs")){
		    		parameters.setInputType(Parameters.COGRAPH_INPUT);
		    	}
		    	else if (input.endsWith(".triples")){
		    		parameters.setInputType(Parameters.TRIPLE_INPUT);
		    	}
		    	else {
		    		parameters.setInputType(Parameters.DEFAULT_INPUT);
		    	}
		    }
		    
		    if( line.hasOption( "output" ) ) {
		    	String output = line.getOptionValue("output");
		    	parameters.setOutput(output);
		    	try {
					new File(output).getCanonicalPath();
				} catch (IOException e) {
					System.out.println( "ERROR: no valid output file specified" );
					showHelp(options);
				}
		    }
		    
		    if( line.hasOption( "cograph-size" ) ) {
		    	int cographSize = ((Number)line.getParsedOptionValue("cograph-size")).intValue();
		    	parameters.setCographSize(cographSize);
		    }
		    
		    if( line.hasOption( "cograph-distance" ) ) {
		    	double cographDistance = ((Number)line.getParsedOptionValue("cograph-distance")).doubleValue();
		    	parameters.setCographDistance(cographDistance);
		    	if (cographDistance<0 || cographDistance>1){
		    		System.out.println( "ERROR: cograph-distance parameter has to be between 0 and 1" );
					showHelp(options);
		    	}
		    }
		    
		    if( line.hasOption( "cograph-all" ) ) {
		    	boolean cographAll = true;
		    	parameters.setCographAllSolutions(cographAll);
		    }
		    
		    if( line.hasOption( "write" ) ) {
		    	boolean write = true;
		    	parameters.setWrite(write);
		    }
		    
		    if( line.hasOption( "enrichment" ) ) {
		    	double enrichment = ((Number)line.getParsedOptionValue("enrichment")).doubleValue();
		    	parameters.setEnrichment(enrichment);
		    	if (enrichment<0 || enrichment>1){
		    		System.out.println( "ERROR: enrichment parameter has to be between 0 and 1" );
		    		showHelp(options);
		    	}
		    	if (parameters.getInputType()!=Parameters.DEFAULT_INPUT) {
					System.out.println( "ERROR: enrichment not possible with triple or cograph input" );
					showHelp(options);
				}
		    }
		    
		    if( line.hasOption( "enrichment-factor" ) ) {
		    	double enrichmentFactor = ((Number)line.getParsedOptionValue("enrichment-factor")).doubleValue();
	    		parameters.setEnrichmentFactor(enrichmentFactor);
	    		if (enrichmentFactor<0){
		    		System.out.println( "ERROR: enrichment-factor parameter has to be greater or equal to 0" );
		    		showHelp(options);
		    	}
		    }
		    
		    if( line.hasOption( "bootstrap" ) ) {
		    	int bootstrap = ((Number)line.getParsedOptionValue("bootstrap")).intValue();
		    	parameters.setBootstrap(bootstrap);
		    	if (bootstrap<0 || bootstrap>2){
		    		System.out.println( "ERROR: bootstrap parameter has to be between 0 and 4" );
					showHelp(options);
		    	}
		    	if (parameters.getInputType()==Parameters.TRIPLE_INPUT && bootstrap==Parameters.COGRAPH_BOOTSTRAPPING) {
					System.out.println( "ERROR: bootstrapping on cographs not possible with triple input" );
					showHelp(options);
				}
		    }
		    
		    if( line.hasOption( "replications" ) ) {
		    	int replications = ((Number)line.getParsedOptionValue("replications")).intValue();
		    	parameters.setReplications(replications);
		    }
		    
		    if( line.hasOption( "support" ) ) {
		    	double support = ((Number)line.getParsedOptionValue("support")).doubleValue();
		    	parameters.setSupport(support);
		    	if (support<0 || support>1){
		    		System.out.println( "ERROR: support parameter has to be between 0 and 1" );
					showHelp(options);
		    	}
		    }

		    if( line.hasOption( "noise-type" ) ) {
		    	int noiseType = ((Number)line.getParsedOptionValue("noise-type")).intValue();
		    	if (noiseType<0 || noiseType>4){
		    		noiseType=0;
		    	}
		    	parameters.setNoiseType(noiseType);
		    }
		    
		    if( line.hasOption( "noise-probability" ) ) {
		    	double noiseProbability = ((Number)line.getParsedOptionValue("noise-probability")).doubleValue();
		    	parameters.setNoiseProbability(noiseProbability);
		    	if (noiseProbability<0 || noiseProbability>1){
		    		System.out.println( "ERROR: noise-probability parameter has to be between 0 and 1" );
					showHelp(options);
		    	}
		    }
		    		    
		    if( line.hasOption( "timelimit-cge" ) ) {
		    	int timelimitCGE = ((Number)line.getParsedOptionValue("timelimit-cge")).intValue();
		    	parameters.setTimelimitCGE(timelimitCGE);
		    }
		    
		    if( line.hasOption( "timelimit-mcs" ) ) {
		    	int timelimitMCS = ((Number)line.getParsedOptionValue("timelimit-mcs")).intValue();
		    	parameters.setTimelimitCGE(timelimitMCS);
		    }
		    
		    if( line.hasOption( "timelimit-lrt" ) ) {
		    	int timelimitLRT = ((Number)line.getParsedOptionValue("timelimit-lrt")).intValue();
		    	parameters.setTimelimitCGE(timelimitLRT);
		    }
		    if( line.hasOption( "quick-te" ) ) {
		    	boolean quickTE = false;
		    	parameters.setQuickTE(quickTE);
		    }
		    if( line.hasOption( "greedy-cs" ) ) {
		    	boolean greedyCS = true;
		    	parameters.setGreedyCS(greedyCS);
		    }
		    
		    if( line.hasOption( "no-ltr" ) ) {
		    	boolean noLTR = true;
		    	parameters.setNoLTR(noLTR);
		    }
		    
		    if( line.hasOption( "verbose" ) ) {
		    	boolean verbose = true;
		    	parameters.setVerbose(verbose);
		    }
		    
		}
		catch( Exception e ) {
		    System.out.println( e.getMessage() );
		    showHelp(options);
		}
		
		return parameters;
	}
	
	public static <T extends Option> void showHelp(Options options){
		HelpFormatter formatter = new HelpFormatter();
		formatter.setOptionComparator(new Comparator<T>() {
			private static final String OPTS_ORDER = "hiosdaefwbrutpcmlqgnv";
		    public int compare(T o1, T o2) {
		        return OPTS_ORDER.indexOf(o1.getOpt()) - OPTS_ORDER.indexOf(o2.getOpt());
		    }
		});
		formatter.printHelp( "java -jar ParaPhylo.jar -i <input file> -o <output file> [OPTIONS]", options );
		System.exit(0);
	}

}
