package paraphylo.solver.cplex;

import ilog.concert.IloException;
import ilog.concert.IloIntVar;
import ilog.concert.IloNumExpr;
import ilog.concert.IloRange;
import ilog.cplex.IloCplex;
import ilog.cplex.IloCplex.UnknownObjectException;

import java.util.List;

import paraphylo.dao.Parameters;
import paraphylo.dataset.GeneFamily;
import paraphylo.dataset.MultiGeneFamilySpeciesDataSet;
import paraphylo.util.NullOutputStream;


public class CplexCographEditingSolver extends CplexAbstractSolver{
	
	// gene family
	private GeneFamily gf;
	// number of genes
	private int noOfGenes;
	
	// orthology variables
	private IloIntVar[][] E;
	
	private int timelimit;
	
	public CplexCographEditingSolver(MultiGeneFamilySpeciesDataSet ds, GeneFamily gf, Parameters parameters) throws IloException{
		super(ds, parameters);
		this.gf = gf;
		this.noOfGenes = this.gf.getNoOfGenes();
		this.timelimit = parameters.getTimelimitCGE();
	}
	
	public String solve() throws IloException{
		this.solver.setName("CplexCographSolver");
		
		// initialize variables E_x,y
		this.E = new IloIntVar[noOfGenes][noOfGenes];
		for (int x=0; x<noOfGenes; x++){
			for (int y=x+1; y<noOfGenes; y++){
				this.E[x][y]= solver.boolVar("E_"+x+","+y);
			}
		}
		
		//TODO uncomment for cograph editing benchmark
		// E_x,y = 0 for x, y from same species
		for (int species=0; species<noOfSpecies; species++){
			List<Integer> genes = this.gf.getGenesOfSpecies(species);
			for(int x : genes){
				for(int y : genes){
					solver.addEq(0, getE(x,y));
				}	
			}
		}
		
		// set the objective function
		int noOfAdjacencies = (this.noOfGenes*(this.noOfGenes-1))/2;
		IloNumExpr theta0Expr[] = new IloNumExpr[noOfAdjacencies];
		IloNumExpr theta1Expr[] = new IloNumExpr[noOfAdjacencies];
		int i=0;
		for (int x=0; x<noOfGenes; x++){
			for (int y=x+1; y<noOfGenes; y++){
				double theta = this.gf.getOrthologyMatrix()[x][y];
				theta0Expr[i] = solver.prod((1-theta), getE(x,y));
				theta1Expr[i] = solver.prod(theta, solver.diff(1, getE(x,y)));
				i++;
			}
		}
		
		this.objFn = solver.sum(solver.sum(theta0Expr),solver.sum(theta1Expr));
		
		solver.addLe(this.objFn, noOfAdjacencies*this.parameters.getCographDistance());
		
		solver.addMinimize(this.objFn);

		// lazy constraints for each p4 
		for (int w=0; w<this.noOfGenes; w++){
			for (int x=0; x<this.noOfGenes; x++){
				if (w!=x){
					for (int y=0; y<this.noOfGenes; y++){
						if (w!=y && x!=y){
							for (int z=w+1; z<this.noOfGenes; z++){
								if (x!=z && y!=z){
									IloIntVar vars[] = {getE(w,x), getE(x,y), getE(y,z),
											getE(w,y), getE(w,z), getE(x,z) };
									double vals[] = {1,1,1,-1,-1,-1};
									IloRange range = solver.le(solver.scalProd(vals, vars), 2, "P4_"+w+","+x+","+y+","+z);
									solver.addLazyConstraint(range);
								}
							}
						}
					}
				}
			}
		}
		
		if (this.timelimit>0){
			solver.setParam(IloCplex.DoubleParam.TiLim, parameters.getTimelimitCGE());
		}
		
		// solve
		if (!parameters.isVerbose()){
			solver.setOut(new NullOutputStream());
		}

		if (this.parameters.isCographAllSolutions()){
			solver.setParam(IloCplex.DoubleParam.SolnPoolAGap, 0.0);
			solver.setParam(IloCplex.IntParam.SolnPoolIntensity, 4);
			solver.setParam(IloCplex.IntParam.PopulateLim, 100);
			solver.populate();
		}
		
		boolean ret = solver.solve();
		String solution = "none";
		if (ret) {
			solution = getSolution();
		}
		// free memory
		solver.end();
		
		return solution;

	}
	

	public IloIntVar getE(int x, int y){
		return this.E[x<y?x:y][x>y?x:y];
	}
	
	/**
	 * @return the current solution, i.e. the values of xVariables as string
	 * @throws UnknownObjectException
	 * @throws IloException
	 */
	private String getSolution() throws UnknownObjectException, IloException {
		int noSolutions = solver.getSolnPoolNsolns();
		String solution = "";
		double bestObjectiveValue = solver.getBestObjValue();
		this.gf.setCographEditDistance((int) Math.round(bestObjectiveValue));
		for (int solutionId=0; solutionId<noSolutions; solutionId++){
			if (solver.getObjValue(solutionId)<=bestObjectiveValue){
				Boolean[][] cograph = new Boolean[this.noOfGenes][this.noOfGenes];
				solution += "Orthology:\n";
				for (int x=0; x<this.noOfGenes; x++){
					for (int y=0; y<this.noOfGenes; y++){
						double variable = 0;
						if (x!=y){
							variable = solver.getValue(getE(x,y), solutionId);
						}
						cograph[x][y]=(variable>0.5);
						solution+=variable<=0.5?"0 ":"1 ";
					}
					solution+="\n";
				}
				gf.addCograph(cograph);
				solution+="\n\n";
				solution+=solver.getObjValue(solutionId)+"\n";
			}
		}
		return solution;
	}
	
}
