package paraphylo.solver.cplex;

import ilog.concert.IloException;
import ilog.concert.IloIntVar;
import ilog.concert.IloRange;
import ilog.cplex.IloCplex;
import ilog.cplex.IloCplex.UnknownObjectException;

import java.util.ArrayList;
import java.util.List;

import paraphylo.dao.Parameters;
import paraphylo.dao.Triple;
import paraphylo.dataset.MultiGeneFamilySpeciesDataSet;
import paraphylo.util.NullOutputStream;


public class CplexMaximalConsistentSubsetSolver extends CplexAbstractSolver{
	
	// triple variables
	private IloIntVar[][][] T;
	private IloIntVar[] tripleVars;
	private double[] tripleVals;
	// resulting dense triple set
	private List<Triple> denseTriples;
	
	private int timelimit;
	
	public CplexMaximalConsistentSubsetSolver(MultiGeneFamilySpeciesDataSet ds, Parameters parameters) throws IloException{
		super(ds, parameters);
		this.timelimit = parameters.getTimelimitMCS();
	}
	
	public String solve() throws IloException{
		this.solver.setName("CplexTripleSolver");
		
		// initialize variables T_ab|c
		int i=0;
		this.tripleVars = new IloIntVar[ds.getNoOfDistinctTriples()];
		this.tripleVals = new double[ds.getNoOfDistinctTriples()];
		this.T = new IloIntVar[noOfSpecies][noOfSpecies][noOfSpecies];
		for (int a=0; a<noOfSpecies; a++){
			for (int b=a+1; b<noOfSpecies; b++){
				for (int c=0; c<noOfSpecies; c++){
					if (a!=c && b!=c){
						T[a][b][c] = solver.boolVar("T_("+a+","+b+"|"+c+")");
						if (ds.getTriple(a, b, c)>0){
							this.tripleVars[i]=T[a][b][c];
							this.tripleVals[i]=ds.getTriple(a, b, c);
							if (this.parameters.getSupport()>0){
								// ignore triples with small support
								if (ds.getSupport(a, b, c) < this.parameters.getSupport()){
									this.tripleVals[i]=0;
								}
							}
							i++;
						}
					}
				}
			}
		}
		
		// T is dense constraint
		for (int A=0; A<noOfSpecies; A++){
			for (int B=A+1; B<noOfSpecies; B++){
				for (int C=B+1; C<noOfSpecies; C++){
					if (A!=C && B!=C){
						solver.addEq(solver.sum(getT(A,B,C), getT(A,C,B), getT(B,C,A)), 1);
					}
				}
			}
		}
		
		// objective function
		//this.objFn = solver.sum(this.tripleVars);
		this.objFn = solver.scalProd(this.tripleVars, this.tripleVals);
		solver.addMaximize(this.objFn);
		
		// lazy constraints for the decker rules
		for (int A=0; A<this.noOfSpecies; A++){
			for (int B=0; B<this.noOfSpecies; B++){
				if (A!=B){
					for (int C=0; C<this.noOfSpecies; C++){
						if (A!=C && B!=C){
							for (int D=0; D<this.noOfSpecies; D++){
								if (A!=D && B!=D && C!=D){
									// 2*T(AB|C) + 2*T(AC|D) - T(AB|D) - T(BC|D) <= 2
									IloRange expr1 = solver.le(solver.sum(solver.prod(2, getT(A,B,C)),solver.prod(2, getT(A,C,D)),solver.negative(getT(A,B,D)),solver.negative(getT(B,C,D))), 2);
									solver.addLazyConstraint(expr1);
									// 2*T(AB|C) + 2*T(CD|A) - T(AB|D) - T(CD|B) <= 2 
//									IloRange expr2 = solver.le(solver.sum(solver.prod(2, getT(A,B,C)),solver.prod(2, getT(C,D,A)),solver.negative(getT(A,B,D)),solver.negative(getT(C,D,B))), 2);
//									solver.addLazyConstraint(expr2);
									// T(AB|D) + T(AC|D) - T(BC|D) <= 1
//									IloRange expr3 = solver.le(solver.sum(getT(A,B,D),getT(A,C,D),solver.negative(getT(B,C,D))), 1);
//									solver.addLazyConstraint(expr3);
								}
							}
						}
					}
				}
			}
		}
		
		if (this.timelimit>0){
			solver.setParam(IloCplex.DoubleParam.TiLim, timelimit);
		}

		// solve
		if (!parameters.isVerbose()){
			solver.setOut(new NullOutputStream());
		}
		boolean ret = solver.solve();
		String solution = "none";
		if (ret) {
			solution = getSolution();
		}
		// free memory
		solver.end();
		
		return solution;

	}
	
	/**
	 * @return the current solution, i.e. the values of xVariables as string
	 * @throws UnknownObjectException
	 * @throws IloException
	 */
	private String getSolution() throws UnknownObjectException, IloException {
		this.denseTriples = new ArrayList<Triple>();
		String solution="Triples:\n";
		int i=0;
		for (int a=0; a<this.noOfSpecies;a++){
			for (int b=a+1; b<this.noOfSpecies; b++){
				for (int c=0; c<this.noOfSpecies; c++){
					if (a!=c && b!=c){
						double variable = solver.getValue(this.T[a][b][c]);
						if (variable>=0.5){
							denseTriples.add(new Triple(a,b,c));
						}
						if (ds.getTriple(a, b, c)>0 && ds.getSupport(a, b, c)>=this.parameters.getSupport()){
							if (variable>=0.5){
								solution+=a+","+b+"|"+c+"\n";
								ds.setCosistentTriple(a, b, c, true);
								i++;
							}
						}
					}
				}
			}
		}
		solution+=i+" triples\n";
		solution+="\n\n";
		solution+=solver.getObjValue()+"\n";
		return solution;
	}
	
	public List<Triple> getDenseTriples() {
		return denseTriples;
	}

	public IloIntVar getT(int x, int y, int z){
		return this.T[x<y?x:y][x>y?x:y][z];
	}
}