package paraphylo.solver.cplex;

import ilog.concert.IloException;
import ilog.concert.IloIntExpr;
import ilog.concert.IloIntVar;
import ilog.cplex.IloCplex.UnknownObjectException;

import java.util.List;

import paraphylo.dao.Parameters;
import paraphylo.dao.Triple;
import paraphylo.dataset.GeneFamily;
import paraphylo.dataset.MultiGeneFamilySpeciesDataSet;
import paraphylo.util.NullOutputStream;


public class CplexTripleExtractionSolver extends CplexAbstractSolver{
	
	// gene family
	private GeneFamily gf;
	// triple variables
	private IloIntVar[][][] T;
	
	private Boolean[][] cograph;	
	
	public CplexTripleExtractionSolver(MultiGeneFamilySpeciesDataSet ds, GeneFamily gf, Boolean[][] cograph, Parameters parameters) throws IloException{
		super(ds, parameters);
		this.gf = gf;
		this.cograph = cograph;
	}
	
	public String solve() throws IloException{
		this.solver.setName("CplexTripleExtractionSolver");
		
		// initialize variables T_AB|C
		this.T = new IloIntVar[noOfSpecies][noOfSpecies][noOfSpecies];
		for (int A=0; A<noOfSpecies; A++){
			for (int B=A+1; B<noOfSpecies; B++){
				for (int C=0; C<noOfSpecies; C++){
					if (A!=C && B!=C){
						this.T[A][B][C] = solver.boolVar("T_("+A+","+B+"|"+C+")");
					}
				}
			}
		}
		
		// extract species triples case 1 (and 2a)
		for (int A=0; A<noOfSpecies; A++){
			for (int B=A+1; B<noOfSpecies; B++){
				for (int C=B+1; C<noOfSpecies; C++){
					List<Integer> genesX = this.gf.getGenesOfSpecies(A);
					for (int x : genesX){
						List<Integer> genesY = this.gf.getGenesOfSpecies(B);
						for (int y : genesY){
							List<Integer> genesZ = this.gf.getGenesOfSpecies(C);
							for (int z : genesZ){
								double xy = this.cograph[x][y]?1:0;
								double xz = this.cograph[x][z]?1:0;
								double yz = this.cograph[y][z]?1:0;
								solver.addLe(solver.sum(1-xy+xz+yz,solver.negative(getT(A,B,C))), 2);
								solver.addLe(solver.sum(xy+1-xz+yz,solver.negative(getT(A,C,B))), 2);
								solver.addLe(solver.sum(xy+xz+1-yz,solver.negative(getT(B,C,A))), 2);
							}
						}
					}
				}
			}
		}
		// extract species triples case 2b
		for (int A=0; A<noOfSpecies; A++){
			for (int B=A+1; B<noOfSpecies; B++){
				for (int C=0; C<noOfSpecies; C++){
					if (A!=C && B!=C) {
						for (int D=0; D<noOfSpecies; D++){
							if (A!=D && B!=D && C!=D){
								IloIntExpr expr[] = { getT(A,D,C), getT(B,D,C), solver.negative(getT(A,B,C)) };
								solver.addLe(solver.sum(expr), 1);
//								solver.addLazyConstraint(solver.le(solver.sum(expr), 1));
							}
						}
					}
				}
			}
		}
		

		IloIntExpr tripleExpr = solver.constant(0);
		for (int A=0; A<noOfSpecies; A++){
			for (int B=A+1; B<noOfSpecies; B++){
				for (int C=0; C<noOfSpecies; C++){
					if (A!=C && B!=C){
						tripleExpr=solver.sum(tripleExpr, getT(A,B,C));
					}
				}
			}
		}
		
		this.objFn = tripleExpr;
		
		solver.addMinimize(this.objFn);

		// solve
		if (!parameters.isVerbose()){
			solver.setOut(new NullOutputStream());
		}
		boolean ret = solver.solve();
		String solution = "none";
		if (ret) {
			solution = getSolution();
		}
		// free memory
		solver.end();
		
		return solution;

	}
	

	public IloIntVar getT(int x, int y, int z){
		return this.T[x<y?x:y][x>y?x:y][z];
	}
	
	/**
	 * @return the current solution, i.e. the values of xVariables as string
	 * @throws UnknownObjectException
	 * @throws IloException
	 */
	private String getSolution() throws UnknownObjectException, IloException {
		List<Triple> triples = gf.getTriples();
		String solution="Triples:\n";
		int i=0;
		for (int A=0; A<this.noOfSpecies;A++){
			for (int B=A+1; B<this.noOfSpecies; B++){
				for (int C=0; C<this.noOfSpecies; C++){
					if (A!=C && B!=C){
						double variable = solver.getValue(getT(A,B,C), 0);
						if (variable>=0.5){
							solution+=A+","+B+"|"+C+"\n";
							ds.addTriple(A, B, C);
							triples.add(new Triple(A,B,C));
							i++;
						}
					}
				}
			}
		}
		solution+=i+" triples\n";
		solution+="\n\n";
		solution+=solver.getObjValue()+"\n";
		return solution;
	}
	
}
