package paraphylo.converter;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.TreeSet;

import org.apache.commons.compress.archivers.tar.TarArchiveEntry;
import org.apache.commons.compress.archivers.tar.TarArchiveInputStream;
import org.apache.commons.compress.compressors.gzip.GzipCompressorInputStream;
import org.apache.commons.io.FileUtils;
import org.apache.commons.io.IOUtils;

public class ALFVPFileParser {
	
	private static Random random = new Random();

	public static String[] parse(String filename, String[] speciesNames,
			Map<String, Integer> geneFamilyMap, List<GeneFamily> geneFamilies, double xpProbability) throws Exception {

		if (speciesNames == null) {
			throw new Exception("speciesNames has to be initialized");
		}
		if (geneFamilyMap==null){
			throw new Exception("geneFamilyMap has to be initialized");
		}
		if (geneFamilies == null) {
			throw new Exception("geneFamilies has to be initialized");
		}

		// extract VP.tgz file
		String hpFilename = null;
		String ppFilename = null;
		String xpFilename = null;
		TarArchiveInputStream tarInput = new TarArchiveInputStream(
				new GzipCompressorInputStream(new FileInputStream(filename)));
		TarArchiveEntry entry = tarInput.getNextTarEntry();
		while (entry != null) {
			String name = entry.getName().substring(
					entry.getName().lastIndexOf('/') + 1);
			if (name.equals("HP.drw") || name.equals("PP.drw") || name.equals("XP.drw")) {
				File tmpFile = File.createTempFile(name.substring(0, 2) + "_",
						name);
				if (name.equals("HP.drw")) {
					hpFilename = tmpFile.getAbsolutePath();
				} else if (name.equals("PP.drw")) {
					ppFilename = tmpFile.getAbsolutePath();
				} else if (name.equals("XP.drw")) {
					xpFilename = tmpFile.getAbsolutePath();
				}
				byte[] content = new byte[(int) entry.getSize()];
				int offset = 0;
				tarInput.read(content, offset, content.length - offset);
				FileOutputStream outputFile = new FileOutputStream(tmpFile);
				IOUtils.write(content, outputFile);
				outputFile.close();
			}
			entry = tarInput.getNextTarEntry();
		}

		List<TreeSet<Tupel<Integer,Integer>>> homologousGroups = getHomologousGroups(hpFilename);
		removeSingleSets(homologousGroups);
		createGeneFamilies(homologousGroups, speciesNames, geneFamilyMap, geneFamilies);
		TreeSet<Tupel<String,String>> paralogousPairs = getParalogousPairs(ppFilename);
		addParalogousPairs(paralogousPairs, geneFamilyMap, geneFamilies);
		// TODO don't try this at home -> perfect orthology
		TreeSet<Tupel<String,String>> xenologousPairs = getParalogousPairs(xpFilename);
		addXenologousPairs(xenologousPairs, geneFamilyMap, geneFamilies, xpProbability);

		FileUtils.deleteQuietly(new File(hpFilename));
		FileUtils.deleteQuietly(new File(ppFilename));
		return null;
	}

	public static List<TreeSet<Tupel<Integer,Integer>>> getHomologousGroups(String hpFilename)
			throws IOException {
		List<TreeSet<Tupel<Integer,Integer>>> homologousGroups = new ArrayList<TreeSet<Tupel<Integer,Integer>>>();

		BufferedReader br = new BufferedReader(new FileReader(hpFilename));
		String hpline = "";
		String line = null;
		while ((line = br.readLine()) != null) {
			line = line.trim();
			if (line.startsWith("#")) {
				continue;
			}
			if (line.matches("^\\s*HP\\s*:=\\s*.*")) {
				hpline = line.substring(line.indexOf('['));
			} else {
				hpline += line;
			}
		}
		br.close();

		int pos = 0;
		int a[] = new int[3];
		int layer = -1;
		String name = "";
		while (pos < hpline.length()) {
			char chr = hpline.charAt(pos);
			if (chr == '[') {
				layer++;
				if (layer == 3) {
					int species1 = a[0];
					int gene1 = a[2];
					Tupel<Integer,Integer> g1 = new Tupel<Integer,Integer>(gene1, species1);
					boolean found = false;
					for (TreeSet<Tupel<Integer,Integer>> gf : homologousGroups) {
						if (gf.contains(g1)) {
							found = true;
							break;
						}
					}
					if (!found) {
						TreeSet<Tupel<Integer,Integer>> gf = new TreeSet<Tupel<Integer,Integer>>();
						gf.add(g1);
						homologousGroups.add(gf);
					}
				}
			} else if (chr == ']') {
				if (name != "") {
					int species1 = a[0];
					int species2 = a[1];
					int gene1 = a[2];
					int gene2 = ((int) Integer.parseInt(name)) - 1;
					Tupel<Integer,Integer> g1 = new Tupel<Integer,Integer>(gene1, species1);
					Tupel<Integer,Integer> g2 = new Tupel<Integer,Integer>(gene2, species2);
					for (TreeSet<Tupel<Integer,Integer>> gf : homologousGroups) {
						if (gf.contains(g1)) {
							gf.add(g2);
							break;
						}
					}
					name = "";
				}
				if (0 <= layer && layer <= 2) {
					a[layer] = 0;
				}
				layer--;
				if (0 <= layer && layer <= 2) {
					a[layer]++;
				}
			} else if (chr == ',') {
				if (name != "") {
					int species1 = a[0];
					int species2 = a[1];
					int gene1 = a[2];
					int gene2 = ((int) Integer.parseInt(name)) - 1;
					Tupel<Integer,Integer> g1 = new Tupel<Integer,Integer>(gene1, species1);
					Tupel<Integer,Integer> g2 = new Tupel<Integer,Integer>(gene2, species2);
					for (TreeSet<Tupel<Integer,Integer>> gf : homologousGroups) {
						if (gf.contains(g1)) {
							gf.add(g2);
							break;
						}
					}
					name = "";
				}
			} else if (chr >= '0' && chr <= '9') {
				name += chr;
			}
			pos++;
		}
		return homologousGroups;
	}
	
	private static void removeSingleSets(
			List<TreeSet<Tupel<Integer, Integer>>> groups) {
		List<TreeSet<Tupel<Integer, Integer>>> toRemove = new ArrayList<TreeSet<Tupel<Integer, Integer>>>();
		for (TreeSet<Tupel<Integer, Integer>> set : groups){
			if (set.size()<=1){
				toRemove.add(set);
			}
		}
		groups.removeAll(toRemove);
	}

	private static void createGeneFamilies(
			List<TreeSet<Tupel<Integer, Integer>>> homologousGroups,
			String[] speciesNames, Map<String, Integer> geneFamilyMap, List<GeneFamily> geneFamilies) {
		int gfCounter = 0;
		for (TreeSet<Tupel<Integer, Integer>> group : homologousGroups){
			String[] species = new String[group.size()];
			String[] geneNames = new String[group.size()];
			int gCounter = 0;
			for (Tupel<Integer, Integer> gene : group){
				geneNames[gCounter] = "G"+(gene.getValue1()+1)+"_SE"+String.format("%03d", (gene.getValue2()+1));
				species[gCounter] = speciesNames[gene.getValue2()];
				geneFamilyMap.put(geneNames[gCounter], gfCounter);
				gCounter++;
			}
			GeneFamily geneFamily = new GeneFamily(species, geneNames);
			for (int i=0; i<group.size(); i++){
				for (int j=0; j<group.size(); j++){
					if (species[i]!=species[j]){
						geneFamily.getPredictedOrthology()[i][j]=1d;
					}
				}
			}
			geneFamilies.add(geneFamily);
			gfCounter++;
		}
		
	}

	public static TreeSet<Tupel<String,String>> getParalogousPairs(String ppFilename)
			throws IOException {
		TreeSet<Tupel<String,String>> paralogousPairs = new TreeSet<Tupel<String,String>>();

		BufferedReader br = new BufferedReader(new FileReader(ppFilename));
		String hpline = "";
		String line = null;
		while ((line = br.readLine()) != null) {
			line = line.trim();
			if (line.startsWith("#")) {
				continue;
			}
			if (line.matches("^\\s*[PX]P\\s*:=\\s*.*")) {
				hpline = line.substring(line.indexOf('['));
			} else {
				hpline += line;
			}
		}
		br.close();

		int pos = 0;
		int a[] = new int[3];
		int layer = -1;
		String name = "";
		while (pos < hpline.length()) {
			char chr = hpline.charAt(pos);
			if (chr == '[') {
				layer++;
			} else if (chr == ']') {
				if (name != "") {
					int species1 = a[0];
					int species2 = a[1];
					if (species1!=species2){
						int gene1 = a[2];
						int gene2 = (int) Integer.parseInt(name) -1 ;
						String geneName1 = "G"+(gene1+1)+"_SE"+String.format("%03d", species1+1);
						String geneName2 = "G"+(gene2+1)+"_SE"+String.format("%03d", species2+1);
						Tupel<String,String> pair = new Tupel<String,String>(geneName1, geneName2);
						paralogousPairs.add(pair);
					}
					name = "";
				}
				if (0 <= layer && layer <= 2) {
					a[layer] = 0;
				}
				layer--;
				if (0 <= layer && layer <= 2) {
					a[layer]++;
				}
			} else if (chr == ',') {
				if (name != "") {
					int species1 = a[0];
					int species2 = a[1];
					if (species1!=species2){
						int gene1 = a[2];
						int gene2 = (int) Integer.parseInt(name) -1 ;
						String geneName1 = "G"+(gene1+1)+"_SE"+String.format("%03d", species1+1);
						String geneName2 = "G"+(gene2+1)+"_SE"+String.format("%03d", species2+1);
						Tupel<String,String> pair = new Tupel<String,String>(geneName1, geneName2);
						paralogousPairs.add(pair);
					}
					name = "";
				}
			} else if (chr >= '0' && chr <= '9') {
				name += chr;
			}
			pos++;
		}
		return paralogousPairs;
	}
	
	private static void addParalogousPairs(
			TreeSet<Tupel<String, String>> paralogousPairs,
			Map<String, Integer> geneFamilyMap, List<GeneFamily> geneFamilies) {
		for (Tupel<String, String> pair : paralogousPairs){
			int geneFamilyId = geneFamilyMap.get(pair.getValue1());
			GeneFamily geneFamily = geneFamilies.get(geneFamilyId);
			List<String> genes = Arrays.asList(geneFamily.getGenes());
			int i = genes.indexOf(pair.getValue1());
			int j = genes.indexOf(pair.getValue2());
			geneFamily.getPredictedOrthology()[i][j]=0d;
		}
		
	}
	
	private static void addXenologousPairs(
			TreeSet<Tupel<String, String>> paralogousPairs,
			Map<String, Integer> geneFamilyMap, List<GeneFamily> geneFamilies, double probability) {
		for (Tupel<String, String> pair : paralogousPairs){
			if (random.nextDouble()<probability){
				int geneFamilyId = geneFamilyMap.get(pair.getValue1());
				GeneFamily geneFamily = geneFamilies.get(geneFamilyId);
				List<String> genes = Arrays.asList(geneFamily.getGenes());
				int i = genes.indexOf(pair.getValue1());
				int j = genes.indexOf(pair.getValue2());
				geneFamily.getPredictedOrthology()[i][j]=0d;
			}
		}
		
	}
}
